// Copyright 2022 Alibaba Group Holding Limited.
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
#ifndef JAVA_CLASS_H
#define JAVA_CLASS_H

#include <java_class/Boolean.h>
#include <java_class/Byte.h>
#include <java_class/Short.h>
#include <java_class/Integer.h>
#include <java_class/Long.h>
#include <java_class/Float.h>
#include <java_class/Double.h>
#include <java_class/Character.h>
#include <java_class/Number.h>
#include <java_class/Iterable.h>
#include <java_class/Iterator.h>
#include <java_class/Collection.h>
#include <java_class/List.h>
#include <java_class/Map.h>
#include <java_class/Entry.h>
#include <java_class/Exception.h>
#include <java_class/Throwable.h>
#include <java_class/StackTraceElement.h>
#include <java_class/LocalDate.h>
#include <java_class/LocalTime.h>
#include <java_class/LocalDateTime.h>
#include <java_class/Date.h>
#include <java_class/Time.h>
#include <java_class/Timestamp.h>
#include <java_class/BigDecimal.h>
#include <java_class/BigInteger.h>
#include <java_class/Constructor.h>
#include <java_class/Method.h>
#include <java_class/Field.h>
#include <java_class/Member.h>
#include <java_class/Modifier.h>
#include <java_class/Class.h>
#include <java_class/Object.h>
#include <java_class/PyIterator.h>
#include <java_class/PyObject.h>

#endif
